import binascii
import threading
from abc import ABC
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Tuple

from yamcs.client.core.exceptions import TimeoutError, YamcsError
from yamcs.client.core.helpers import parse_server_time, parse_value, to_isostring
from yamcs.client.model import Event
from yamcs.protobuf.alarms import alarms_pb2
from yamcs.protobuf.pvalue import pvalue_pb2

__all__ = [
    "Acknowledgment",
    "Alarm",
    "AlarmRangeSet",
    "AlarmUpdate",
    "Calibrator",
    "CommandHistory",
    "ContainerData",
    "EventAlarm",
    "IssuedCommand",
    "MonitoredCommand",
    "Packet",
    "ParameterAlarm",
    "ParameterData",
    "ParameterValue",
    "ValueUpdate",
    "VerificationConfig",
]


def _parse_alarm(proto):
    """Converts a protobuf alarm to a specific Alarm implementation."""
    if proto.type == alarms_pb2.PARAMETER:
        return ParameterAlarm(proto)
    if proto.type == alarms_pb2.EVENT:
        return EventAlarm(proto)
    raise YamcsError("Unexpected type " + str(proto.type))


class Acknowledgment:
    def __init__(self, name, time, status, message):
        self.name = name
        """Name of this acknowledgment."""

        self.time = time
        """Last update time of this acknowledgment."""

        self.status = status
        """Status of this acknowlegment."""

        self.message = message
        """Message of this acknowledgment."""

    def is_terminated(self):
        return self.status and self.status != "SCHEDULED" and self.status != "PENDING"

    def __repr__(self):
        return f"{self.name}: {self.status}"

    def __str__(self):
        return self.__repr__()


class CommandHistory:
    def __init__(self, proto):
        self.generation_time: datetime = parse_server_time(proto.generationTime)
        """The generation time as set by Yamcs"""

        self._proto = proto
        self.attributes = {}
        self._update(proto.attr)

    @property
    def id(self) -> str:
        """
        A unique identifier for this command.
        """
        return self._proto.id

    @property
    def name(self) -> str:
        """Name of the command."""
        return self._proto.commandName

    @property
    def aliases(self) -> Dict[str, str]:
        """
        Aliases, keyed by namespace

        .. versionadded:: 1.9.2
           Compatible with Yamcs 5.8.4 onwards
        """
        return {key: self._proto.aliases[key] for key in self._proto.aliases}

    @property
    def origin(self) -> Optional[str]:
        """
        The origin of this command. Usually the IP address of the issuer.
        """
        if self._proto.HasField("origin"):
            return self._proto.origin
        return None

    @property
    def sequence_number(self) -> int:
        """
        The sequence number of this command. This is the sequence
        number assigned by the issuing client.
        """
        return self._proto.sequenceNumber

    @property
    def username(self) -> Optional[str]:
        """Username of the issuer."""
        return self.attributes.get("username")

    @property
    def assignments(self) -> Dict[str, Any]:
        """
        Argument assignments by name.

        This returns only explicit assignments set by the command
        issuer. To return all assignments, including ones that
        use their default values, use :attr:`all_assignments` instead.

        .. versionadded:: 1.9.2
        """
        args = {}
        for assignment in self._proto.assignments:
            if assignment.userInput:
                args[assignment.name] = parse_value(assignment.value)
        return args

    @property
    def all_assignments(self) -> Dict[str, Any]:
        """
        Argument assignments by name.

        .. versionadded:: 1.9.2
        """
        args = {}
        for assignment in self._proto.assignments:
            args[assignment.name] = parse_value(assignment.value)
        return args

    @property
    def source(self) -> str:
        """String representation of the command."""
        result = self.name + "("
        args = []
        for assignment in self._proto.assignments:
            if assignment.userInput:
                value = parse_value(assignment.value)
                if isinstance(value, str):
                    value = '"' + value + '"'
                elif isinstance(value, bytes) or isinstance(value, bytearray):
                    value = '"' + binascii.hexlify(value).decode("ascii") + '"'
                elif isinstance(value, datetime):
                    value = '"' + to_isostring(value) + '"'
                args.append(assignment.name + ": " + str(value))
        result += ", ".join(args)
        result += ")"
        return result

    @property
    def unprocessed_binary(self):
        """
        Binary representation before postprocessing.

        .. versionadded:: 1.9.2
        """
        if self._proto.HasField("unprocessedBinary"):
            return self._proto.unprocessedBinary
        return None

    @property
    def binary(self):
        """Binary representation of the command."""
        return self.attributes.get("binary")

    def is_complete(self) -> bool:
        """
        Returns whether this command is complete. A command
        can be completed, yet still failed.
        """
        ack = self._assemble_ack("CommandComplete")
        return (ack is not None) and (ack.status == "OK" or ack.status == "NOK")

    def is_success(self) -> bool:
        """
        Returns True if the command has completed successfully.
        """
        ack = self._assemble_ack("CommandComplete")
        return (ack is not None) and ack.status == "OK"

    def is_failure(self) -> bool:
        """
        Returns True if the command has completed, but failed.
        """
        ack = self._assemble_ack("CommandComplete")
        return (ack is not None) and ack.status == "NOK"

    @property
    def error(self) -> Optional[str]:
        """Error message in case the command failed."""
        ack = self._assemble_ack("CommandComplete")
        if ack and ack.status == "NOK":
            return ack.message
        return None

    @property
    def comment(self) -> Optional[str]:
        """Optional user comment attached when issuing the command."""
        return self.attributes.get("comment")

    @property
    def acknowledgments(self) -> Dict[str, Acknowledgment]:
        """
        All acknowledgments by name.

        :return:
            Acknowledgments keyed by name.
        """
        acks = {}
        for name, _ in self.attributes.items():
            if name.startswith("CommandComplete"):
                continue
            if name.startswith("TransmissionConstraints"):
                continue
            if name.endswith("_Status"):
                ack = self._assemble_ack(name[:-7])
                if ack:
                    acks[ack.name] = ack

        return acks

    def _assemble_ack(self, name: str) -> Optional[Acknowledgment]:
        time = self.attributes.get(name + "_Time")
        status = self.attributes.get(name + "_Status")
        message = self.attributes.get(name + "_Message")
        if time and status:
            return Acknowledgment(name, time, status, message)
        return None

    def _update(self, proto):
        # Copy-on-write to avoid mutation error on the exposed attributes field
        copy = self.attributes.copy()
        for attr in proto:
            value = parse_value(attr.value)
            copy[attr.name] = value
        self.attributes = copy

    def __str__(self):
        acks = ", ".join(ack.__repr__() for _, ack in self.acknowledgments.items())
        return f"{self.name} [{acks}]"


class IssuedCommand:
    def __init__(self, proto):
        self._proto = proto

    @property
    def id(self) -> str:
        """
        A unique identifier for this command.
        """
        return self._proto.id

    @property
    def name(self) -> str:
        """
        The fully-qualified name of this command.
        """
        return self._proto.commandName

    @property
    def aliases(self) -> Dict[str, str]:
        """
        Aliases, keyed by namespace

        .. versionadded:: 1.9.2
           Compatible with Yamcs 5.8.4 onwards
        """
        return {key: self._proto.aliases[key] for key in self._proto.aliases}

    @property
    def generation_time(self) -> datetime:
        """
        The generation time as set by Yamcs.
        """
        return parse_server_time(self._proto.generationTime)

    @property
    def username(self) -> str:
        """The username of the issuer."""
        return self._proto.username

    @property
    def assignments(self) -> Dict[str, Any]:
        """
        Argument assignments by name.

        This returns only explicit assignments set by the command
        issuer. To return all assignments, including ones that
        use their default values, use :attr:`all_assignments` instead.

        .. versionadded:: 1.9.2
        """
        args = {}
        for assignment in self._proto.assignments:
            if assignment.userInput:
                args[assignment.name] = parse_value(assignment.value)
        return args

    @property
    def all_assignments(self) -> Dict[str, Any]:
        """
        Argument assignments by name.

        .. versionadded:: 1.9.2
        """
        args = {}
        for assignment in self._proto.assignments:
            args[assignment.name] = parse_value(assignment.value)
        return args

    @property
    def queue(self) -> Optional[str]:
        """The name of the queue that this command was assigned to."""
        if self._proto.HasField("queue"):
            return self._proto.queue
        return None

    @property
    def origin(self) -> Optional[str]:
        """
        The origin of this command. Usually the IP address of the issuer.
        """
        if self._proto.HasField("origin"):
            return self._proto.origin
        return None

    @property
    def sequence_number(self):
        """
        The sequence number of this command. This is the sequence
        number assigned by the issuing client.
        """
        if self._proto.HasField("sequenceNumber"):
            return self._proto.sequenceNumber
        return None

    @property
    def source(self):
        """String representation of this command."""
        result = None
        if self.name:
            result = self.name + "("
            args = []
            for assignment in self._proto.assignments:
                if assignment.userInput:
                    value = parse_value(assignment.value)
                    if isinstance(value, str):
                        value = '"' + value + '"'
                    elif isinstance(value, bytes) or isinstance(value, bytearray):
                        value = '"' + binascii.hexlify(value).decode("ascii") + '"'
                    elif isinstance(value, datetime):
                        value = '"' + to_isostring(value) + '"'
                    args.append(assignment.name + ": " + str(value))
            result += ", ".join(args)
            result += ")"

        return result

    @property
    def hex(self):
        """Hexadecimal string representation of this command."""
        if self._proto.HasField("binary"):
            return binascii.hexlify(self._proto.binary).decode("ascii")
        return None

    @property
    def unprocessed_binary(self):
        """Binary representation before postprocessing."""
        if self._proto.HasField("unprocessedBinary"):
            return self._proto.unprocessedBinary
        return None

    @property
    def binary(self):
        """Binary representation of this command."""
        if self._proto.HasField("binary"):
            return self._proto.binary
        return None

    def __str__(self):
        return f"{self.generation_time} {self.source}"


class VerificationConfig:
    """
    Contains overrides to the default verification handling of Yamcs.
    """

    def __init__(self):
        self._disabled = []
        self._disable_all = False
        self._check_windows = {}

    def disable(self, verifier: Optional[str] = None):
        """
        Disable verification.

        :param verifier:
            Name of a specific verifier to disable. If unspecified
            all verifiers are disabled.
        """
        if verifier:
            self._disabled.append(verifier)
        else:
            self._disable_all = True

    def modify_check_window(
        self, verifier: str, start: Optional[float] = None, stop: Optional[float] = None
    ):
        """
        Set or override the check window.

        Depending on the Mission Database configuration,
        the time may be relative to either the command release
        or a preceding verifier.

        :param verifier:
            Name of the verifier
        :param start:
            Window start time (relative, in seconds)
        :param stop:
            Window stop time (relative, in seconds)
        """
        self._check_windows[verifier] = {"start": start, "stop": stop}


class MonitoredCommand(IssuedCommand):
    """
    Represent an instance of an issued command that is updated
    throughout the acknowledgment process.

    Objects of this class are owned by a :class:`.CommandConnection` instance.
    """

    def __init__(self, proto):
        super(MonitoredCommand, self).__init__(proto)
        self._cmdhist = None
        self._completed = threading.Event()
        self._ack_events = {}

    def _process_cmdhist(self, cmdhist):
        self._cmdhist = cmdhist
        if self.is_complete():
            self._completed.set()
        for name, ack in self.acknowledgments.items():
            event = self._ack_events.setdefault(name, threading.Event())
            if ack.is_terminated():
                event.set()

    @property
    def attributes(self) -> Dict[str, Any]:
        if self._cmdhist:
            return self._cmdhist.attributes
        return {}

    def is_complete(self) -> bool:
        """
        Returns whether this command is complete. A command
        can be completed, yet still failed.

        Use :meth:`await_complete` to wait until this information
        is available.
        """
        ack = self._assemble_ack("CommandComplete")
        return (ack is not None) and (ack.status == "OK" or ack.status == "NOK")

    def is_success(self) -> bool:
        """
        Returns True if this command was completed successfully.

        Use :meth:`await_complete` to wait until this information
        is available.
        """
        ack = self._assemble_ack("CommandComplete")
        return (ack is not None) and ack.status == "OK"

    def is_failure(self) -> bool:
        """
        Returns True if the command has completed, but failed.

        Use :meth:`await_complete` to wait until this information
        is available.
        """
        ack = self._assemble_ack("CommandComplete")
        return (ack is not None) and ack.status == "NOK"

    def await_complete(self, timeout: Optional[float] = None):
        """
        Wait for the command to be `completed`. Afterwards use
        :meth:`is_success` or :meth:`is_failure` to determine
        whether the command was successful or not.

        .. note::

            Yamcs cannot determine completion unless the command has
            an appropriate verifier configured in the Yamcs MDB.

            When no such verifier is present, this method will never
            return (or timeout).

        :param timeout:
            The amount of seconds to wait.
        """
        self._wait_on_signal(self._completed, timeout)

    def await_acknowledgment(
        self, name: str, timeout: Optional[float] = None
    ) -> Acknowledgment:
        """
        Waits for the result of a specific acknowledgment.

        :param name:
            The name of the acknowledgment. Standard names are
            ``Acknowledge_Queued``, ``Acknowledge_Released``
            and ``Acknowledge_Sent``. Others depend on
            specific link types.
        :param timeout:
            The amount of seconds to wait.
        """
        event = self._ack_events.setdefault(name, threading.Event())
        self._wait_on_signal(event, timeout)
        ack = self.acknowledgments.get(name)
        assert ack is not None
        return ack

    def _wait_on_signal(self, event: threading.Event, timeout: Optional[float] = None):
        if not event.wait(timeout=timeout):
            # Remark that a timeout does *not* mean that the underlying
            # work is canceled.
            raise TimeoutError("Timed out.")

    @property
    def error(self) -> Optional[str]:
        """
        Error message in case the command failed.

        This information is only available when the command has failed
        to complete. (:meth:`is_failure` returns `True`).
        """
        ack = self._assemble_ack("CommandComplete")
        if ack and ack.status == "NOK":
            return ack.message
        return None

    @property
    def comment(self) -> Optional[str]:
        """Optional user comment attached when issuing the command."""
        return self.attributes.get("comment")

    @property
    def acknowledgments(self) -> Dict[str, Acknowledgment]:
        """
        All acknowledgments by name.

        :return:
            Acknowledgments keyed by name.
        """
        acks = {}
        for name, _ in self.attributes.items():
            if name.startswith("CommandComplete"):
                continue
            if name.startswith("TransmissionConstraints"):
                continue
            if name.endswith("_Status"):
                ack = self._assemble_ack(name[:-7])
                if ack:
                    acks[ack.name] = ack

        return acks

    def _assemble_ack(self, name: str) -> Optional[Acknowledgment]:
        time = self.attributes.get(name + "_Time")
        status = self.attributes.get(name + "_Status")
        message = self.attributes.get(name + "_Message")
        if time and status:
            return Acknowledgment(name, time, status, message)
        return None

    def _update(self, proto):
        # Copy-on-write to avoid mutation error on the exposed attributes field
        copy = self.attributes.copy()
        for attr in proto:
            value = parse_value(attr.value)
            copy[attr.name] = value
        self.attributes = copy

    def __str__(self):
        acks = ", ".join(ack.__repr__() for _, ack in self.acknowledgments.items())
        return f"{self.name} [{acks}]"


class AlarmUpdate:
    """
    Object received through callbacks when subscribing to alarm updates.
    """

    def __init__(self, proto):
        self._proto = proto

    @property
    def update_type(self) -> str:
        """Type of update."""
        return alarms_pb2.AlarmNotificationType.Name(self._proto.type)

    @property
    def alarm(self) -> "Alarm":
        """
        Latest alarm state.
        """
        return _parse_alarm(self._proto)

    def __str__(self):
        return f"[{self.update_type}] {self.alarm}"


class Alarm(ABC):
    def __init__(self, proto):
        self._proto = proto

    @property
    def name(self) -> str:
        """Fully-qualified name of the source of this alarm."""
        if self._proto.id.HasField("namespace"):
            return self._proto.id.namespace + "/" + self._proto.id.name
        return self._proto.id.name

    @property
    def trigger_time(self) -> datetime:
        """
        Processor time when the alarm was triggered.
        """
        return parse_server_time(self._proto.triggerTime)

    @property
    def update_time(self) -> datetime:
        """
        Processor time when the alarm was last updated.
        """
        if self._proto.HasField("updateTime"):
            return parse_server_time(self._proto.updateTime)
        return None

    @property
    def severity(self) -> str:
        return alarms_pb2.AlarmSeverity.Name(self._proto.severity)

    @property
    def sequence_number(self) -> int:
        """
        Sequence number for this specific alarm instance. This allows ensuring
        that operations (such as acknowledgment) are done on the expected alarm
        instance.
        """
        return self._proto.seqNum

    @property
    def is_ok(self) -> bool:
        """
        True if this alarm is currently 'inactive'.

        For a non-latching alarm this is identical to :meth:`is_process_ok`. A
        latching alarm is only OK if it has returned to normal and was
        acknowledged.
        """
        return not self._proto.triggered

    @property
    def is_process_ok(self) -> bool:
        """
        True if the process that caused this alarm is OK. For example:
        parameter back within limits.

        For a non-latching alarm this is identical to :meth:`is_ok`.
        """
        return self._proto.processOK

    @property
    def is_latching(self) -> bool:
        """
        True if this is a latching alarm. A latching alarm returns to
        normal only when the operator resets it
        """
        return self._proto.HasField("latching") and self._proto.latching

    @property
    def is_latched(self) -> bool:
        """
        True if this alarm is currently latched.
        """
        return self.is_latching and self.is_process_ok and not self.is_ok

    @property
    def is_acknowledged(self) -> bool:
        """True if this alarm has been acknowledged."""
        return self._proto.acknowledged

    @property
    def acknowledged_by(self) -> Optional[str]:
        """Username of the acknowledger."""
        if self.is_acknowledged and self._proto.acknowledgeInfo.HasField(
            "acknowledgedBy"
        ):
            return self._proto.acknowledgeInfo.acknowledgedBy
        return None

    @property
    def acknowledge_message(self) -> Optional[str]:
        """Comment provided when acknowledging the alarm."""
        if self.is_acknowledged and self._proto.acknowledgeInfo.HasField(
            "acknowledgeMessage"
        ):
            return self._proto.acknowledgeInfo.acknowledgeMessage
        return None

    @property
    def acknowledge_time(self) -> Optional[datetime]:
        """
        Processor time when the alarm was acknowledged.
        """
        if self.is_acknowledged and self._proto.acknowledgeInfo.HasField(
            "acknowledgeTime"
        ):
            return parse_server_time(self._proto.acknowledgeInfo.acknowledgeTime)
        return None

    @property
    def is_shelved(self) -> bool:
        """True if this alarm has been shelved."""
        return self._proto.HasField("shelveInfo")

    @property
    def violation_count(self) -> int:
        """
        Number of violating samples while this alarm is active.
        """
        return self._proto.violations

    @property
    def count(self) -> int:
        """
        Total number of samples while this alarm is active.
        """
        return self._proto.count

    def __str__(self):
        return f"{self.name} ({self.violation_count} violations)"


class ParameterAlarm(Alarm):
    """
    An alarm triggered by a parameter that went out of limits.
    """

    @property
    def trigger_value(self) -> "ParameterValue":
        """
        Parameter value that originally triggered the alarm
        """
        return ParameterValue(self._proto.parameterDetail.triggerValue)

    @property
    def most_severe_value(self) -> "ParameterValue":
        """
        First parameter value that invoked the highest severity
        level of this alarm.
        """
        return ParameterValue(self._proto.parameterDetail.mostSevereValue)

    @property
    def current_value(self) -> "ParameterValue":
        """
        Latest parameter value for this alarm.
        """
        return ParameterValue(self._proto.parameterDetail.currentValue)


class EventAlarm(Alarm):
    """
    An alarm triggered by an event.
    """

    @property
    def trigger_event(self) -> Event:
        """
        Event that originally triggered the alarm
        """
        return Event(self._proto.eventDetail.triggerEvent)

    @property
    def most_severe_event(self) -> Event:
        """
        First event that invoked the highest severity level
        of this alarm
        """
        return Event(self._proto.eventDetail.mostSevereEvent)

    @property
    def current_event(self) -> Event:
        """
        Latest event for this alarm
        """
        return Event(self._proto.eventDetail.currentEvent)


class ParameterValue:
    def __init__(self, proto, id=None):
        self._proto = proto
        self._id = id or proto.id

    @property
    def name(self) -> str:
        """
        An identifying name for the parameter value. Typically this is the
        fully-qualified XTCE name, but it may also be an alias depending
        on how the parameter update was requested.
        """
        if self._id.namespace:
            return self._id.namespace + "/" + self._id.name
        return self._id.name

    @property
    def generation_time(self) -> datetime:
        """
        The time when the parameter was generated. If the parameter
        was extracted from a packet, this usually returns the packet time.
        """
        return parse_server_time(self._proto.generationTime)

    @property
    def reception_time(self) -> Optional[datetime]:
        """
        The time when the parameter value was received by Yamcs.
        """
        if self._proto.HasField("acquisitionTime"):
            return parse_server_time(self._proto.acquisitionTime)
        return None

    @property
    def validity_duration(self) -> Optional[timedelta]:
        """
        How long this parameter value is valid.

        .. note: There is also an option when subscribing to get updated when
                 the parameter values expire.
        """
        if self._proto.HasField("expireMillis"):
            return timedelta(milliseconds=self._proto.expireMillis)
        return None

    @property
    def raw_value(self):
        """
        The raw (uncalibrated) value.
        """
        if self._proto.HasField("rawValue"):
            return parse_value(self._proto.rawValue)
        return None

    @property
    def eng_value(self):
        """
        The engineering (calibrated) value.
        """
        if self._proto.HasField("engValue"):
            return parse_value(self._proto.engValue)
        return None

    @property
    def monitoring_result(self) -> str:
        if self._proto.HasField("monitoringResult"):
            return pvalue_pb2.MonitoringResult.Name(self._proto.monitoringResult)
        return None

    @property
    def range_condition(self) -> Optional[str]:
        """
        If the value is out of limits, this indicates ``LOW`` or ``HIGH``.
        """
        if self._proto.HasField("rangeCondition"):
            return pvalue_pb2.RangeCondition.Name(self._proto.rangeCondition)
        return None

    @property
    def validity_status(self) -> str:
        if self._proto.HasField("acquisitionStatus"):
            return pvalue_pb2.AcquisitionStatus.Name(self._proto.acquisitionStatus)
        return None

    def __str__(self):
        line = f"{self.generation_time} {self.name} {self.raw_value} {self.eng_value}"
        if self.monitoring_result:
            line += " [" + self.monitoring_result + "]"
        return line


class ParameterData:
    def __init__(self, proto, mapping=None):
        self._proto = proto
        self._mapping = mapping

        self._pvals = {}
        if self._mapping is None:  # ParameterData proto
            for pval_pb in proto.parameter:
                pval = ParameterValue(pval_pb)
                self._pvals[pval.name] = pval
        else:  # SubscribeParametersData proto
            for pval_pb in self._proto.values:
                id = self._mapping[pval_pb.numericId]
                pval = ParameterValue(pval_pb, id=id)
                self._pvals[pval.name] = pval

    def __iter__(self):
        for pval in self._pvals.values():
            yield pval

    def get_value(self, parameter: str) -> Optional[ParameterValue]:
        """
        Returns the value of a specific parameter.
        Or ``None`` if the parameter is not included in
        this update.

        :param parameter:
            Parameter name.
        """
        return self._pvals.get(parameter)

    @property
    def parameters(self) -> List[ParameterValue]:
        return list(self._pvals.values())

    def __str__(self):
        return self.parameters.__str__()


class Calibrator:
    """
    A calibrator that may be applied to a numeric raw value.

    Two types of calibrators can be applied:

    * Polynomial calibrators apply a polynomial expression of the form:
        `y = a + bx + cx^2 + ...`.

        The `data` argument must be an array of floats ``[a, b, c, ...]``.

    * Spline calibrators interpolate the raw value between a set of points
        which represent a linear curve.

        The `data` argument must be an array of ``[x, y]`` points.
    """

    POLYNOMIAL = "polynomial"
    SPLINE = "spline"

    def __init__(self, context: str, type: str, data):
        """
        :param context:
            Condition under which this calibrator may be
            applied. The value ``None`` indicates the
            default calibrator which is only applied if
            no contextual calibrators match.
        :param type:
            One of ``polynomial`` or ``spline``.
        :param data:
            Calibration definition for the selected type.
        """
        self.context = context
        self.type = type
        self.data = data


class ContainerData:
    def __init__(self, proto):
        self._proto = proto

    @property
    def name(self) -> str:
        """
        The name of the container.
        """
        return self._proto.name

    @property
    def generation_time(self) -> datetime:
        """
        The time when this container's packet was generated (packet time).
        """
        return parse_server_time(self._proto.generationTime)

    @property
    def reception_time(self) -> datetime:
        """
        The time when this container's packet was received by Yamcs.
        """
        return parse_server_time(self._proto.receptionTime)

    @property
    def binary(self):
        """
        Raw binary
        """
        if self._proto.HasField("binary"):
            return self._proto.binary
        return None

    def __str__(self):
        return f"{self.generation_time} ({self.name})"


class Packet:
    def __init__(self, proto):
        self._proto = proto

    @property
    def name(self) -> str:
        """
        The name of the packet. When using XTCE extraction this is the
        fully-qualified name of the first container in the hierarchy that
        this packet maps to.
        """
        return self._proto.id.name

    @property
    def generation_time(self) -> datetime:
        """
        The time when the packet was generated (packet time).
        """
        return parse_server_time(self._proto.generationTime)

    @property
    def earth_reception_time(self) -> Optional[datetime]:
        """
        When the signal was received on the ground.

        .. versionadded:: 1.9.1
        """
        if self._proto.HasField("earthReceptionTime"):
            return parse_server_time(self._proto.earthReceptionTime)
        return None

    @property
    def reception_time(self) -> datetime:
        """
        The time when the packet was received by Yamcs.
        """
        return parse_server_time(self._proto.receptionTime)

    @property
    def sequence_number(self) -> int:
        """
        The sequence number of the packet. This is usually decoded from
        the packet.
        """
        return self._proto.sequenceNumber

    @property
    def link(self) -> str:
        """
        Name of the Yamcs link where this packet was received from.

        .. versionadded:: 1.9.1
        """
        return self._proto.link

    @property
    def binary(self) -> bytes:
        """
        Raw binary of this packet
        """
        if self._proto.HasField("packet"):
            return self._proto.packet
        return b""

    @property
    def size(self) -> int:
        """
        Size in bytes of this packet
        """
        if self._proto.HasField("size"):
            return self._proto.size
        return len(self.binary)

    def __str__(self):
        return f"{self.generation_time} #{self.sequence_number} ({self.name})"


class AlarmRangeSet:
    """
    A set of alarm ranges that apply in a specific context.
    """

    def __init__(
        self,
        context: str,
        watch: Optional[Tuple[Optional[float], Optional[float]]] = None,
        warning: Optional[Tuple[Optional[float], Optional[float]]] = None,
        distress: Optional[Tuple[Optional[float], Optional[float]]] = None,
        critical: Optional[Tuple[Optional[float], Optional[float]]] = None,
        severe: Optional[Tuple[Optional[float], Optional[float]]] = None,
        min_violations: int = 1,
    ):
        """
        :param context:
            Condition under which this range set is
            applicable. The value ``None`` indicates the
            default range set which is only applicable if
            no contextual sets match.
        :param watch:
            Range expressed as a tuple ``(lo, hi)``
            where lo and hi are assumed exclusive.
        :param warning:
            Range expressed as a tuple ``(lo, hi)``
            where lo and hi are assumed exclusive.
        :param distress:
            Range expressed as a tuple ``(lo, hi)``
            where lo and hi are assumed exclusive.
        :param critical:
            Range expressed as a tuple ``(lo, hi)``
            where lo and hi are assumed exclusive.
        :param severe:
            Range expressed as a tuple ``(lo, hi)``
            where lo and hi are assumed exclusive.
        :param min_violations:
            Minimum violations before an alarm is
            generated.
        """
        self.context = context
        self.watch = watch
        self.warning = warning
        self.distress = distress
        self.critical = critical
        self.severe = severe
        self.min_violations = min_violations


class ValueUpdate:
    """
    Data holder for passing a value along with its generation time when
    updating a software parameter.
    """

    def __init__(
        self,
        value: Any,
        generation_time: Optional[datetime] = None,
        expires_in: Optional[float] = None,
    ):
        """
        :param value:
            The value to set
        :param generation_time:
            Generation time of the value. If unset, Yamcs will
            assign the generation time.
        :param expires_in:
            How long before this value expires (in fractional seconds).
            If unset, the value does not expire.

            .. versionadded:: 1.9.1
               Compatible with Yamcs 5.8.8 onwards
        """
        self.value = value
        self.generation_time = generation_time
        self.expires_in = expires_in



import warnings

warnings.warn(
    "The module 'yamcs.filetransfer.model' is deprecated. "
    "Import classes from 'yamcs.client' instead.",
    DeprecationWarning,
    stacklevel=2,
)

from yamcs.client.filetransfer.model import (  # noqa
    EntityInfo,
    FileTransferCapabilities,
    FileTransferOption,
    FileTransferService,
    RemoteFile,
    RemoteFileListing,
    Transfer,
)

Service = FileTransferService
"""
Temporary backwards compatibility.
Prefer to use the class 'FileTransferService'.
"""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: yamcs/protobuf/links/links.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from yamcs.api import annotations_pb2 as yamcs_dot_api_dot_annotations__pb2
from yamcs.protobuf.actions import actions_pb2 as yamcs_dot_protobuf_dot_actions_dot_actions__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='yamcs/protobuf/links/links.proto',
  package='yamcs.protobuf.links',
  syntax='proto2',
  serialized_options=b'\n\030org.yamcs.protobuf.linksB\021LinksServiceProtoP\001',
  serialized_pb=b'\n yamcs/protobuf/links/links.proto\x12\x14yamcs.protobuf.links\x1a\x1byamcs/api/annotations.proto\x1a$yamcs/protobuf/actions/actions.proto\x1a\x1cgoogle/protobuf/struct.proto\"$\n\x10ListLinksRequest\x12\x10\n\x08instance\x18\x01 \x01(\t\"B\n\x11ListLinksResponse\x12-\n\x05links\x18\x01 \x03(\x0b\x32\x1e.yamcs.protobuf.links.LinkInfo\"0\n\x0eGetLinkRequest\x12\x10\n\x08instance\x18\x01 \x01(\t\x12\x0c\n\x04link\x18\x02 \x01(\t\"3\n\x11\x45nableLinkRequest\x12\x10\n\x08instance\x18\x01 \x01(\t\x12\x0c\n\x04link\x18\x02 \x01(\t\"4\n\x12\x44isableLinkRequest\x12\x10\n\x08instance\x18\x01 \x01(\t\x12\x0c\n\x04link\x18\x02 \x01(\t\":\n\x18ResetLinkCountersRequest\x12\x10\n\x08instance\x18\x01 \x01(\t\x12\x0c\n\x04link\x18\x02 \x01(\t\"l\n\x10RunActionRequest\x12\x10\n\x08instance\x18\x01 \x01(\t\x12\x0c\n\x04link\x18\x02 \x01(\t\x12\x0e\n\x06\x61\x63tion\x18\x03 \x01(\t\x12(\n\x07message\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\":\n\tLinkEvent\x12-\n\x05links\x18\x03 \x03(\x0b\x32\x1e.yamcs.protobuf.links.LinkInfo\")\n\x15SubscribeLinksRequest\x12\x10\n\x08instance\x18\x01 \x01(\t\"\xbc\x02\n\x08LinkInfo\x12\x10\n\x08instance\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x0c\n\x04spec\x18\x04 \x01(\t\x12\x10\n\x08\x64isabled\x18\x06 \x01(\x08\x12\x0e\n\x06status\x18\x07 \x01(\t\x12\x13\n\x0b\x64\x61taInCount\x18\n \x01(\x03\x12\x14\n\x0c\x64\x61taOutCount\x18\x0b \x01(\x03\x12\x16\n\x0e\x64\x65tailedStatus\x18\t \x01(\t\x12\x12\n\nparentName\x18\x0c \x01(\t\x12\x33\n\x07\x61\x63tions\x18\r \x03(\x0b\x32\".yamcs.protobuf.actions.ActionInfo\x12&\n\x05\x65xtra\x18\x0e \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x12\n\nparameters\x18\x0f \x03(\tJ\x04\x08\x05\x10\x06J\x04\x08\x08\x10\t2\xfa\x07\n\x08LinksApi\x12y\n\tListLinks\x12&.yamcs.protobuf.links.ListLinksRequest\x1a\'.yamcs.protobuf.links.ListLinksResponse\"\x1b\x8a\x92\x03\x17\n\x15/api/links/{instance}\x12s\n\x07GetLink\x12$.yamcs.protobuf.links.GetLinkRequest\x1a\x1e.yamcs.protobuf.links.LinkInfo\"\"\x8a\x92\x03\x1e\n\x1c/api/links/{instance}/{link}\x12\x97\x01\n\nEnableLink\x12\'.yamcs.protobuf.links.EnableLinkRequest\x1a\x1e.yamcs.protobuf.links.LinkInfo\"@\x8a\x92\x03<\x1a#/api/links/{instance}/{link}:enableb\x15Link \'{link}\' enabled\x12\x9b\x01\n\x0b\x44isableLink\x12(.yamcs.protobuf.links.DisableLinkRequest\x1a\x1e.yamcs.protobuf.links.LinkInfo\"B\x8a\x92\x03>\x1a$/api/links/{instance}/{link}:disableb\x16Link \'{link}\' disabled\x12\x95\x01\n\x11ResetLinkCounters\x12..yamcs.protobuf.links.ResetLinkCountersRequest\x1a\x1e.yamcs.protobuf.links.LinkInfo\"0\x8a\x92\x03,\x1a*/api/links/{instance}/{link}:resetCounters\x12m\n\x0eSubscribeLinks\x12+.yamcs.protobuf.links.SubscribeLinksRequest\x1a\x1f.yamcs.protobuf.links.LinkEvent\"\x0b\xda\x92\x03\x07\n\x05links0\x01\x12\xb8\x01\n\tRunAction\x12&.yamcs.protobuf.links.RunActionRequest\x1a\x17.google.protobuf.Struct\"j\x8a\x92\x03\x66\x1a-/api/links/{instance}/{link}/actions/{action}:\x07messageb,Action \'{action}\' performed on link \'{link}\'\x1a\x04\xd0\x80\x01\x01\x42/\n\x18org.yamcs.protobuf.linksB\x11LinksServiceProtoP\x01'
  ,
  dependencies=[yamcs_dot_api_dot_annotations__pb2.DESCRIPTOR,yamcs_dot_protobuf_dot_actions_dot_actions__pb2.DESCRIPTOR,google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])




_LISTLINKSREQUEST = _descriptor.Descriptor(
  name='ListLinksRequest',
  full_name='yamcs.protobuf.links.ListLinksRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance', full_name='yamcs.protobuf.links.ListLinksRequest.instance', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=155,
  serialized_end=191,
)


_LISTLINKSRESPONSE = _descriptor.Descriptor(
  name='ListLinksResponse',
  full_name='yamcs.protobuf.links.ListLinksResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='links', full_name='yamcs.protobuf.links.ListLinksResponse.links', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=193,
  serialized_end=259,
)


_GETLINKREQUEST = _descriptor.Descriptor(
  name='GetLinkRequest',
  full_name='yamcs.protobuf.links.GetLinkRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance', full_name='yamcs.protobuf.links.GetLinkRequest.instance', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='link', full_name='yamcs.protobuf.links.GetLinkRequest.link', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=261,
  serialized_end=309,
)


_ENABLELINKREQUEST = _descriptor.Descriptor(
  name='EnableLinkRequest',
  full_name='yamcs.protobuf.links.EnableLinkRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance', full_name='yamcs.protobuf.links.EnableLinkRequest.instance', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='link', full_name='yamcs.protobuf.links.EnableLinkRequest.link', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=311,
  serialized_end=362,
)


_DISABLELINKREQUEST = _descriptor.Descriptor(
  name='DisableLinkRequest',
  full_name='yamcs.protobuf.links.DisableLinkRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance', full_name='yamcs.protobuf.links.DisableLinkRequest.instance', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='link', full_name='yamcs.protobuf.links.DisableLinkRequest.link', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=364,
  serialized_end=416,
)


_RESETLINKCOUNTERSREQUEST = _descriptor.Descriptor(
  name='ResetLinkCountersRequest',
  full_name='yamcs.protobuf.links.ResetLinkCountersRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance', full_name='yamcs.protobuf.links.ResetLinkCountersRequest.instance', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='link', full_name='yamcs.protobuf.links.ResetLinkCountersRequest.link', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=418,
  serialized_end=476,
)


_RUNACTIONREQUEST = _descriptor.Descriptor(
  name='RunActionRequest',
  full_name='yamcs.protobuf.links.RunActionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance', full_name='yamcs.protobuf.links.RunActionRequest.instance', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='link', full_name='yamcs.protobuf.links.RunActionRequest.link', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='action', full_name='yamcs.protobuf.links.RunActionRequest.action', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='yamcs.protobuf.links.RunActionRequest.message', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=478,
  serialized_end=586,
)


_LINKEVENT = _descriptor.Descriptor(
  name='LinkEvent',
  full_name='yamcs.protobuf.links.LinkEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='links', full_name='yamcs.protobuf.links.LinkEvent.links', index=0,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=588,
  serialized_end=646,
)


_SUBSCRIBELINKSREQUEST = _descriptor.Descriptor(
  name='SubscribeLinksRequest',
  full_name='yamcs.protobuf.links.SubscribeLinksRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance', full_name='yamcs.protobuf.links.SubscribeLinksRequest.instance', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=648,
  serialized_end=689,
)


_LINKINFO = _descriptor.Descriptor(
  name='LinkInfo',
  full_name='yamcs.protobuf.links.LinkInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance', full_name='yamcs.protobuf.links.LinkInfo.instance', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='yamcs.protobuf.links.LinkInfo.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='yamcs.protobuf.links.LinkInfo.type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spec', full_name='yamcs.protobuf.links.LinkInfo.spec', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='disabled', full_name='yamcs.protobuf.links.LinkInfo.disabled', index=4,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='status', full_name='yamcs.protobuf.links.LinkInfo.status', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dataInCount', full_name='yamcs.protobuf.links.LinkInfo.dataInCount', index=6,
      number=10, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dataOutCount', full_name='yamcs.protobuf.links.LinkInfo.dataOutCount', index=7,
      number=11, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='detailedStatus', full_name='yamcs.protobuf.links.LinkInfo.detailedStatus', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parentName', full_name='yamcs.protobuf.links.LinkInfo.parentName', index=9,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='actions', full_name='yamcs.protobuf.links.LinkInfo.actions', index=10,
      number=13, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extra', full_name='yamcs.protobuf.links.LinkInfo.extra', index=11,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='yamcs.protobuf.links.LinkInfo.parameters', index=12,
      number=15, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=692,
  serialized_end=1008,
)

_LISTLINKSRESPONSE.fields_by_name['links'].message_type = _LINKINFO
_RUNACTIONREQUEST.fields_by_name['message'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_LINKEVENT.fields_by_name['links'].message_type = _LINKINFO
_LINKINFO.fields_by_name['actions'].message_type = yamcs_dot_protobuf_dot_actions_dot_actions__pb2._ACTIONINFO
_LINKINFO.fields_by_name['extra'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
DESCRIPTOR.message_types_by_name['ListLinksRequest'] = _LISTLINKSREQUEST
DESCRIPTOR.message_types_by_name['ListLinksResponse'] = _LISTLINKSRESPONSE
DESCRIPTOR.message_types_by_name['GetLinkRequest'] = _GETLINKREQUEST
DESCRIPTOR.message_types_by_name['EnableLinkRequest'] = _ENABLELINKREQUEST
DESCRIPTOR.message_types_by_name['DisableLinkRequest'] = _DISABLELINKREQUEST
DESCRIPTOR.message_types_by_name['ResetLinkCountersRequest'] = _RESETLINKCOUNTERSREQUEST
DESCRIPTOR.message_types_by_name['RunActionRequest'] = _RUNACTIONREQUEST
DESCRIPTOR.message_types_by_name['LinkEvent'] = _LINKEVENT
DESCRIPTOR.message_types_by_name['SubscribeLinksRequest'] = _SUBSCRIBELINKSREQUEST
DESCRIPTOR.message_types_by_name['LinkInfo'] = _LINKINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ListLinksRequest = _reflection.GeneratedProtocolMessageType('ListLinksRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTLINKSREQUEST,
  '__module__' : 'yamcs.protobuf.links.links_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.links.ListLinksRequest)
  })
_sym_db.RegisterMessage(ListLinksRequest)

ListLinksResponse = _reflection.GeneratedProtocolMessageType('ListLinksResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTLINKSRESPONSE,
  '__module__' : 'yamcs.protobuf.links.links_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.links.ListLinksResponse)
  })
_sym_db.RegisterMessage(ListLinksResponse)

GetLinkRequest = _reflection.GeneratedProtocolMessageType('GetLinkRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETLINKREQUEST,
  '__module__' : 'yamcs.protobuf.links.links_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.links.GetLinkRequest)
  })
_sym_db.RegisterMessage(GetLinkRequest)

EnableLinkRequest = _reflection.GeneratedProtocolMessageType('EnableLinkRequest', (_message.Message,), {
  'DESCRIPTOR' : _ENABLELINKREQUEST,
  '__module__' : 'yamcs.protobuf.links.links_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.links.EnableLinkRequest)
  })
_sym_db.RegisterMessage(EnableLinkRequest)

DisableLinkRequest = _reflection.GeneratedProtocolMessageType('DisableLinkRequest', (_message.Message,), {
  'DESCRIPTOR' : _DISABLELINKREQUEST,
  '__module__' : 'yamcs.protobuf.links.links_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.links.DisableLinkRequest)
  })
_sym_db.RegisterMessage(DisableLinkRequest)

ResetLinkCountersRequest = _reflection.GeneratedProtocolMessageType('ResetLinkCountersRequest', (_message.Message,), {
  'DESCRIPTOR' : _RESETLINKCOUNTERSREQUEST,
  '__module__' : 'yamcs.protobuf.links.links_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.links.ResetLinkCountersRequest)
  })
_sym_db.RegisterMessage(ResetLinkCountersRequest)

RunActionRequest = _reflection.GeneratedProtocolMessageType('RunActionRequest', (_message.Message,), {
  'DESCRIPTOR' : _RUNACTIONREQUEST,
  '__module__' : 'yamcs.protobuf.links.links_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.links.RunActionRequest)
  })
_sym_db.RegisterMessage(RunActionRequest)

LinkEvent = _reflection.GeneratedProtocolMessageType('LinkEvent', (_message.Message,), {
  'DESCRIPTOR' : _LINKEVENT,
  '__module__' : 'yamcs.protobuf.links.links_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.links.LinkEvent)
  })
_sym_db.RegisterMessage(LinkEvent)

SubscribeLinksRequest = _reflection.GeneratedProtocolMessageType('SubscribeLinksRequest', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIBELINKSREQUEST,
  '__module__' : 'yamcs.protobuf.links.links_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.links.SubscribeLinksRequest)
  })
_sym_db.RegisterMessage(SubscribeLinksRequest)

LinkInfo = _reflection.GeneratedProtocolMessageType('LinkInfo', (_message.Message,), {
  'DESCRIPTOR' : _LINKINFO,
  '__module__' : 'yamcs.protobuf.links.links_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.links.LinkInfo)
  })
_sym_db.RegisterMessage(LinkInfo)


DESCRIPTOR._options = None

_LINKSAPI = _descriptor.ServiceDescriptor(
  name='LinksApi',
  full_name='yamcs.protobuf.links.LinksApi',
  file=DESCRIPTOR,
  index=0,
  serialized_options=b'\320\200\001\001',
  serialized_start=1011,
  serialized_end=2029,
  methods=[
  _descriptor.MethodDescriptor(
    name='ListLinks',
    full_name='yamcs.protobuf.links.LinksApi.ListLinks',
    index=0,
    containing_service=None,
    input_type=_LISTLINKSREQUEST,
    output_type=_LISTLINKSRESPONSE,
    serialized_options=b'\212\222\003\027\n\025/api/links/{instance}',
  ),
  _descriptor.MethodDescriptor(
    name='GetLink',
    full_name='yamcs.protobuf.links.LinksApi.GetLink',
    index=1,
    containing_service=None,
    input_type=_GETLINKREQUEST,
    output_type=_LINKINFO,
    serialized_options=b'\212\222\003\036\n\034/api/links/{instance}/{link}',
  ),
  _descriptor.MethodDescriptor(
    name='EnableLink',
    full_name='yamcs.protobuf.links.LinksApi.EnableLink',
    index=2,
    containing_service=None,
    input_type=_ENABLELINKREQUEST,
    output_type=_LINKINFO,
    serialized_options=b'\212\222\003<\032#/api/links/{instance}/{link}:enableb\025Link \'{link}\' enabled',
  ),
  _descriptor.MethodDescriptor(
    name='DisableLink',
    full_name='yamcs.protobuf.links.LinksApi.DisableLink',
    index=3,
    containing_service=None,
    input_type=_DISABLELINKREQUEST,
    output_type=_LINKINFO,
    serialized_options=b'\212\222\003>\032$/api/links/{instance}/{link}:disableb\026Link \'{link}\' disabled',
  ),
  _descriptor.MethodDescriptor(
    name='ResetLinkCounters',
    full_name='yamcs.protobuf.links.LinksApi.ResetLinkCounters',
    index=4,
    containing_service=None,
    input_type=_RESETLINKCOUNTERSREQUEST,
    output_type=_LINKINFO,
    serialized_options=b'\212\222\003,\032*/api/links/{instance}/{link}:resetCounters',
  ),
  _descriptor.MethodDescriptor(
    name='SubscribeLinks',
    full_name='yamcs.protobuf.links.LinksApi.SubscribeLinks',
    index=5,
    containing_service=None,
    input_type=_SUBSCRIBELINKSREQUEST,
    output_type=_LINKEVENT,
    serialized_options=b'\332\222\003\007\n\005links',
  ),
  _descriptor.MethodDescriptor(
    name='RunAction',
    full_name='yamcs.protobuf.links.LinksApi.RunAction',
    index=6,
    containing_service=None,
    input_type=_RUNACTIONREQUEST,
    output_type=google_dot_protobuf_dot_struct__pb2._STRUCT,
    serialized_options=b'\212\222\003f\032-/api/links/{instance}/{link}/actions/{action}:\007messageb,Action \'{action}\' performed on link \'{link}\'',
  ),
])
_sym_db.RegisterServiceDescriptor(_LINKSAPI)

DESCRIPTOR.services_by_name['LinksApi'] = _LINKSAPI

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: yamcs/api/exception.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='yamcs/api/exception.proto',
  package='yamcs.api',
  syntax='proto3',
  serialized_options=b'\n\rorg.yamcs.apiB\016ExceptionProtoP\001',
  serialized_pb=b'\n\x19yamcs/api/exception.proto\x12\tyamcs.api\x1a\x19google/protobuf/any.proto\"a\n\x10\x45xceptionMessage\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x0b\n\x03msg\x18\x03 \x01(\t\x12$\n\x06\x64\x65tail\x18\x04 \x01(\x0b\x32\x14.google.protobuf.Any\"c\n\x15\x46ilterSyntaxException\x12\x11\n\tbeginLine\x18\x01 \x01(\x05\x12\x13\n\x0b\x62\x65ginColumn\x18\x02 \x01(\x05\x12\x0f\n\x07\x65ndLine\x18\x03 \x01(\x05\x12\x11\n\tendColumn\x18\x04 \x01(\x05\x42!\n\rorg.yamcs.apiB\x0e\x45xceptionProtoP\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])




_EXCEPTIONMESSAGE = _descriptor.Descriptor(
  name='ExceptionMessage',
  full_name='yamcs.api.ExceptionMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='yamcs.api.ExceptionMessage.code', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='yamcs.api.ExceptionMessage.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='msg', full_name='yamcs.api.ExceptionMessage.msg', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='detail', full_name='yamcs.api.ExceptionMessage.detail', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=67,
  serialized_end=164,
)


_FILTERSYNTAXEXCEPTION = _descriptor.Descriptor(
  name='FilterSyntaxException',
  full_name='yamcs.api.FilterSyntaxException',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='beginLine', full_name='yamcs.api.FilterSyntaxException.beginLine', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beginColumn', full_name='yamcs.api.FilterSyntaxException.beginColumn', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='endLine', full_name='yamcs.api.FilterSyntaxException.endLine', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='endColumn', full_name='yamcs.api.FilterSyntaxException.endColumn', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=166,
  serialized_end=265,
)

_EXCEPTIONMESSAGE.fields_by_name['detail'].message_type = google_dot_protobuf_dot_any__pb2._ANY
DESCRIPTOR.message_types_by_name['ExceptionMessage'] = _EXCEPTIONMESSAGE
DESCRIPTOR.message_types_by_name['FilterSyntaxException'] = _FILTERSYNTAXEXCEPTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ExceptionMessage = _reflection.GeneratedProtocolMessageType('ExceptionMessage', (_message.Message,), {
  'DESCRIPTOR' : _EXCEPTIONMESSAGE,
  '__module__' : 'yamcs.api.exception_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.api.ExceptionMessage)
  })
_sym_db.RegisterMessage(ExceptionMessage)

FilterSyntaxException = _reflection.GeneratedProtocolMessageType('FilterSyntaxException', (_message.Message,), {
  'DESCRIPTOR' : _FILTERSYNTAXEXCEPTION,
  '__module__' : 'yamcs.api.exception_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.api.FilterSyntaxException)
  })
_sym_db.RegisterMessage(FilterSyntaxException)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import functools
import urllib.parse
import warnings
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from yamcs.client.archive.model import (
    IndexGroup,
    MeanSample,
    ParameterRange,
    ResultSet,
    Stream,
    StreamData,
    Table,
)
from yamcs.client.core import pagination
from yamcs.client.core.context import Context
from yamcs.client.core.exceptions import YamcsError
from yamcs.client.core.futures import WebSocketSubscriptionFuture
from yamcs.client.core.helpers import (
    split_protobuf_stream,
    to_isostring,
    to_named_object_ids,
    to_server_time,
)
from yamcs.client.core.subscriptions import WebSocketSubscriptionManager
from yamcs.client.model import Event
from yamcs.client.tmtc.model import (
    Alarm,
    CommandHistory,
    EventAlarm,
    Packet,
    ParameterAlarm,
    ParameterData,
    ParameterValue,
)
from yamcs.protobuf.alarms import alarms_pb2, alarms_service_pb2
from yamcs.protobuf.archive import index_service_pb2, parameter_archive_service_pb2
from yamcs.protobuf.commanding import commanding_pb2, commands_service_pb2
from yamcs.protobuf.events import events_pb2, events_service_pb2
from yamcs.protobuf.packets import packets_pb2, packets_service_pb2
from yamcs.protobuf.pvalue import pvalue_pb2, pvalue_service_pb2
from yamcs.protobuf.table import table_pb2

__all__ = [
    "ArchiveClient",
]


def _wrap_callback_parse_stream_data(subscription, on_data, message):
    """
    Wraps an (optional) user callback to parse StreamData
    from a WebSocket data message
    """
    pb = table_pb2.StreamData()
    message.Unpack(pb)
    on_data(StreamData(pb))


class ArchiveClient:
    def __init__(self, ctx: Context, instance: str):
        super(ArchiveClient, self).__init__()
        self.ctx = ctx
        self._instance = instance

    def list_packet_names(self) -> Iterable[str]:
        """
        Returns the existing packet names.
        """
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        path = f"/archive/{self._instance}/packet-names"
        response = self.ctx.get_proto(path=path)
        message = packets_service_pb2.ListPacketNamesResponse()
        message.ParseFromString(response.content)
        names = getattr(message, "name")
        return iter(names)

    def list_packet_histogram(
        self,
        name: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        merge_time: float = 2,
    ) -> Iterable[IndexGroup]:
        """
        Reads packet-related index records between the specified start and stop
        time.

        Each iteration returns a chunk of chronologically-sorted records.

        :param merge_time:
            Maximum gap in seconds before two consecutive index
            records are merged together.
        """
        params = {}
        if name is not None:
            params["name"] = name
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if merge_time is not None:
            params["mergeTime"] = int(merge_time * 1000)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/packet-index",
            params=params,
            response_class=index_service_pb2.IndexResponse,
            items_key="group",
            item_mapper=IndexGroup,
        )

    def list_processed_parameter_groups(self) -> Iterable[str]:
        """
        Returns the existing parameter groups.
        """
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        path = f"/archive/{self._instance}/parameter-groups"
        response = self.ctx.get_proto(path=path)
        message = pvalue_service_pb2.ListParameterGroupsResponse()
        message.ParseFromString(response.content)
        groups = getattr(message, "groups")
        return iter(groups)

    def list_processed_parameter_group_histogram(
        self,
        group: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        merge_time: float = 20,
    ) -> Iterable[IndexGroup]:
        """
        Reads index records related to processed parameter groups between the
        specified start and stop time.

        Each iteration returns a chunk of chronologically-sorted records.

        :param merge_time:
            Maximum gap in seconds before two consecutive index
            records are merged together.
        """
        params = {}
        if group is not None:
            params["group"] = group
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if merge_time is not None:
            params["mergeTime"] = int(merge_time * 1000)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/parameter-index",
            params=params,
            response_class=index_service_pb2.IndexResponse,
            items_key="group",
            item_mapper=IndexGroup,
        )

    def list_event_sources(self) -> Iterable[str]:
        """
        Returns the existing event sources.
        """
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        path = f"/archive/{self._instance}/events/sources"
        response = self.ctx.get_proto(path=path)
        message = events_service_pb2.ListEventSourcesResponse()
        message.ParseFromString(response.content)
        sources = getattr(message, "source")
        return iter(sources)

    def list_event_histogram(
        self,
        source: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        merge_time: float = 2,
    ) -> Iterable[IndexGroup]:
        """
        Reads event-related index records between the specified start and stop
        time.

        Each iteration returns a chunk of chronologically-sorted records.

        :param merge_time:
            Maximum gap in seconds before two consecutive index
            records are merged together.
        """
        params = {}
        if source is not None:
            params["source"] = source
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if merge_time is not None:
            params["mergeTime"] = int(merge_time * 1000)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/event-index",
            params=params,
            response_class=index_service_pb2.IndexResponse,
            items_key="group",
            item_mapper=IndexGroup,
        )

    def list_command_histogram(
        self,
        name: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        merge_time: float = 2,
    ) -> Iterable[IndexGroup]:
        """
        Reads command-related index records between the specified start and stop
        time.

        Each iteration returns a chunk of chronologically-sorted records.

        :param merge_time:
            Maximum gap in seconds before two consecutive index
            records are merged together.
        """
        params = {}
        if name is not None:
            params["name"] = name
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if merge_time is not None:
            params["mergeTime"] = int(merge_time * 1000)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/command-index",
            params=params,
            response_class=index_service_pb2.IndexResponse,
            items_key="group",
            item_mapper=IndexGroup,
        )

    def list_completeness_index(
        self,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        merge_time: float = 2,
    ) -> Iterable[IndexGroup]:
        """
        Reads completeness index records between the specified start and stop
        time.

        Each iteration returns a chunk of chronologically-sorted records.

        :param merge_time:
            Maximum gap in seconds before two consecutive index
            records are merged together.
        """
        params = {}
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if merge_time is not None:
            params["mergeTime"] = int(merge_time * 1000)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/completeness-index",
            params=params,
            response_class=index_service_pb2.IndexResponse,
            items_key="group",
            item_mapper=IndexGroup,
        )

    def list_alarms(
        self,
        name: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        page_size: int = 500,
        descending: bool = False,
    ) -> Iterable[Alarm]:
        """
        Reads alarm information between the specified start and stop time.

        Alarms are sorted by trigger time, name and sequence number.

        :param name:
            Filter by alarm name
        :param start:
            Minimum trigger time (inclusive)
        :param stop:
            Maximum trigger time (exclusive)
        :param page_size:
            Page size of underlying requests. Higher values imply
            less overhead, but risk hitting the maximum message size
            limit.
        :param descending:
            If set to ``True`` alarms are fetched in reverse
            order (most recent first).
        """
        params: Dict[str, Any] = {
            "order": "desc" if descending else "asc",
        }
        if name is not None:
            params["name"] = name
        if page_size is not None:
            params["limit"] = page_size
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        def to_alarm(proto):
            if proto.type == alarms_pb2.PARAMETER:
                return ParameterAlarm(proto)
            elif proto.type == alarms_pb2.EVENT:
                return EventAlarm(proto)
            else:
                raise YamcsError("Unexpected type " + str(proto.type))

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/alarms",
            params=params,
            response_class=alarms_service_pb2.ListAlarmsResponse,
            items_key="alarms",
            item_mapper=to_alarm,
        )

    def list_packets(
        self,
        name: Optional[str] = None,
        filter: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        page_size: int = 500,
        descending: bool = False,
    ) -> Iterable[Packet]:
        """
        Reads packet information between the specified start and stop
        time.

        Packets are sorted by generation time and sequence number.

        .. note::

            This method will send out multiple requests when more than
            ``page_size`` packets are queried. For large queries, consider
            using :meth:`stream_packets` instead, it uses server-streaming
            based on a single request.

        :param name:
            Archived name of the packet
        :param filter:
            Filter query, allows for both text and field search.

            .. versionadded:: 1.12.0
               Compatible with Yamcs v5.12.0 onwards
        :param start:
            Minimum generation time of the returned packets (inclusive)
        :param stop:
            Maximum generation time of the returned packets (exclusive)
        :param page_size:
            Page size of underlying requests. Higher values imply
            less overhead, but risk hitting the maximum message size
            limit.
        :param descending:
            If set to ``True`` packets are fetched in reverse
            order (most recent first).
        """
        params: Dict[str, Any] = {
            "order": "desc" if descending else "asc",
        }
        if name is not None:
            params["name"] = name
        if page_size is not None:
            params["limit"] = page_size
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if filter is not None:
            params["filter"] = filter

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/packets",
            params=params,
            response_class=packets_service_pb2.ListPacketsResponse,
            items_key="packet",
            item_mapper=Packet,
        )

    def stream_packets(
        self,
        name: Optional[Union[str, List[str]]] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        chunk_size: int = 32 * 1024,
    ) -> Iterable[Packet]:
        """
        Reads packet information between the specified start and stop time.

        .. versionadded:: 1.9.1

        :param name:
            If specified, return only packets with this archived name.
        :param start:
            Minimum generation time of the returned packets (inclusive)
        :param stop:
            Maximum generation time of the returned packets (exclusive)
        """
        options = packets_service_pb2.StreamPacketsRequest()
        if name is not None:
            if isinstance(name, str):
                options.name.extend([name])
            else:
                options.name.extend(name)
        if start is not None:
            options.start.MergeFrom(to_server_time(start))
        if stop is not None:
            options.stop.MergeFrom(to_server_time(stop))

        def generate():
            path = f"/stream-archive/{self._instance}:streamPackets"
            response = self.ctx.post_proto(
                path=path, data=options.SerializeToString(), stream=True
            )
            for message in split_protobuf_stream(
                response.iter_content(chunk_size=chunk_size), packets_pb2.TmPacketData
            ):
                yield Packet(message)

        return generate()

    def get_packet(
        self,
        generation_time: datetime,
        sequence_number: int,
        partition: Optional[str] = None,
    ) -> Packet:
        """
        Gets a single packet by its identifying key (partition, gentime, seqNum).

        :param generation_time:
            When the packet was generated (packet time)
        :param sequence_number:
            Sequence number of the packet
        :param partition:
            Packet partition name. This property works only against recent versions
            of Yamcs,and will become required in the future (it was erroneously
            omitted).
        """
        url = f"/archive/{self._instance}/packets/"

        # Currently we allow the partition to be unspecified, but this should
        # become deprecated within a few more Yamcs versions.
        # (this endpoint variation was only recently introduced)
        if partition:
            url += urllib.parse.quote_plus(partition) + "/"
        url += f"{to_isostring(generation_time)}/{sequence_number}"
        response = self.ctx.get_proto(url)
        message = packets_pb2.TmPacketData()
        message.ParseFromString(response.content)
        return Packet(message)

    def export_parameter_values(
        self,
        parameters: List[str],
        *,
        namespace: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        interval: Optional[float] = None,
        chunk_size: int = 32 * 1024,
        source: str = "ParameterArchive",
    ) -> Iterable:
        """
        Export parameter values in CSV format.

        :param parameters:
            List of parameter names. These may be
            fully-qualified XTCE name or an alias
            in the format ``NAMESPACE/NAME``.
        :param namespace:
            Preferred namespace of the names in the returned CSV header
        :param start:
            Minimum generation time of the returned values (inclusive)
        :param stop:
            Maximum generation time of the returned values (exclusive)
        :param interval:
            If specified, only one value for each interval is returned. The interval is
            expressed in seconds.
        :param source:
            Specify how to retrieve parameter values. By
            default this uses the ``ParameterArchive`` which
            is optimized for parameter retrieval. For Yamcs
            instances that do not enable the ``ParameterArchive``,
            you can still get results by specifying ``replay`` as
            the source. Replay requests take longer to return
            because the packet data needs to be reprocessed.

            .. versionadded:: 1.13.0
               Compatible with Yamcs v5.12.6 onwards
        :return:
            An iterator over received chunks
        """
        params: Dict[str, Any] = {
            "parameters": parameters,
            "source": source,
        }
        if namespace is not None:
            params["namespace"] = namespace
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if interval is not None:
            params["interval"] = int(interval * 1000)

        path = f"/archive/{self._instance}:exportParameterValues"
        response = self.ctx.get_proto(path=path, params=params, stream=True)
        return response.iter_content(chunk_size=chunk_size)

    def export_commands(
        self,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        chunk_size: int = 32 * 1024,
    ) -> Iterable:
        """
        Export command history.

        Commands are sorted by generation time, origin and sequence number.

        :param start:
            Minimum generation time of the returned commands (inclusive)
        :param stop:
            Maximum generation time of the returned commands (exclusive)
        :return:
            An iterator over received chunks
        """
        params = {}
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        path = f"/archive/{self._instance}:exportCommands"
        response = self.ctx.get_proto(path=path, params=params, stream=True)
        return response.iter_content(chunk_size=chunk_size)

    def export_packets(
        self,
        name: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        chunk_size: int = 32 * 1024,
    ) -> Iterable:
        """
        Export raw packets.

        Packets are sorted by generation time and sequence number.

        :param name:
            Archived name of the packet
        :param start:
            Minimum generation time of the returned packets (inclusive)
        :param stop:
            Maximum generation time of the returned packets (exclusive)
        :return:
            An iterator over received chunks
        """
        params = {}
        if name is not None:
            params["name"] = name
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        path = f"/archive/{self._instance}:exportPackets"
        response = self.ctx.get_proto(path=path, params=params, stream=True)
        return response.iter_content(chunk_size=chunk_size)

    def list_events(
        self,
        source: Optional[str] = None,
        severity: Optional[str] = None,
        text_filter: Optional[str] = None,
        filter: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        page_size: int = 500,
        descending: bool = False,
    ) -> Iterable[Event]:
        """
        Reads events between the specified start and stop time.

        Events are sorted by generation time, source, then sequence number.

        .. note::

            This method will send out multiple requests when more than
            ``page_size`` events are queried. For large queries, consider
            using :meth:`stream_events` instead, it uses server-streaming
            based on a single request.

        :param source:
            The source of the returned events.
        :param severity:
            The minimum severity level of the returned events.
            One of ``INFO``, ``WATCH``, ``WARNING``, ``DISTRESS``,
            ``CRITICAL`` or ``SEVERE``.
        :param text_filter:
            Filter the text message of the returned events
        :param filter:
            Filter query, allows for both text and field search.

            .. versionadded:: 1.10.1
               Compatible with Yamcs v5.10.2 onwards
        :param start:
            Minimum start date of the returned events (inclusive)
        :param stop:
            Maximum start date of the returned events (exclusive)
        :param page_size:
            Page size of underlying requests. Higher values imply
            less overhead, but risk hitting the maximum message size
            limit.
        :param descending:
            If set to ``True`` events are fetched in reverse
            order (most recent first).
        """
        params: Dict[str, Any] = {
            "order": "desc" if descending else "asc",
        }
        if source is not None:
            params["source"] = source
        if page_size is not None:
            params["limit"] = page_size
        if severity is not None:
            params["severity"] = severity
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if text_filter is not None:
            params["q"] = text_filter
        if filter is not None:
            params["filter"] = filter

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/events",
            params=params,
            response_class=events_service_pb2.ListEventsResponse,
            items_key="event",
            item_mapper=Event,
        )

    def stream_events(
        self,
        source: Optional[Union[str, List[str]]] = None,
        severity: Optional[str] = None,
        text_filter: Optional[str] = None,
        filter: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        chunk_size: int = 32 * 1024,
    ) -> Iterable[Event]:
        """
        Reads events between the specified start and stop time.

        .. versionadded:: 1.9.1

        :param source:
            If specified, return only events with this source.
        :param severity:
            The minimum severity level of the returned events.
            One of ``INFO``, ``WATCH``, ``WARNING``, ``DISTRESS``,
            ``CRITICAL`` or ``SEVERE``.
        :param text_filter:
            Filter the text message of the returned events
        :param filter:
            Filter query applied to returned events. Allows for both text
            and field search.

            .. versionadded:: 1.10.1
               Compatible with Yamcs v5.10.2 onwards
        :param start:
            Minimum generation time of the returned events (inclusive)
        :param stop:
            Maximum generation time of the returned events (exclusive)
        """
        options = events_service_pb2.StreamEventsRequest()
        if source is not None:
            if isinstance(source, str):
                options.source.extend([source])
            else:
                options.source.extend(source)
        if severity is not None:
            options.severity = severity
        if start is not None:
            options.start.MergeFrom(to_server_time(start))
        if stop is not None:
            options.stop.MergeFrom(to_server_time(stop))
        if text_filter is not None:
            options.q = text_filter
        if filter is not None:
            options.filter = filter

        def generate():
            path = f"/stream-archive/{self._instance}:streamEvents"
            response = self.ctx.post_proto(
                path=path, data=options.SerializeToString(), stream=True
            )
            for message in split_protobuf_stream(
                response.iter_content(chunk_size=chunk_size), events_pb2.Event
            ):
                yield Event(message)

        return generate()

    def sample_parameter_values(
        self,
        parameter: str,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        sample_count: int = 500,
        parameter_cache: str = "realtime",
        source: str = "ParameterArchive",
    ) -> List[MeanSample]:
        """
        .. warning::
            Deprecated. Since yamcs-client 1.12.0 this method was
            renamed to :meth:`downsample_mean`.
        """
        warnings.warn(
            "The method 'sample_parameter_values' is deprecated. "
            "Use 'downsample_mean' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.downsample_mean(
            parameter,
            start=start,
            stop=stop,
            sample_count=sample_count,
            parameter_cache=parameter_cache,
            source=source,
        )

    def downsample_mean(
        self,
        parameter: str,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        *,
        sample_count: int = 500,
        parameter_cache: str = "realtime",
        source: str = "ParameterArchive",
    ) -> List[MeanSample]:
        """
        Downsample parameter values in the requested range using
        averaging (mean).

        The query range is split in sample intervals of equal length. For
        each interval a :class:`.MeanSample` is returned which describes
        the min, max, count and avg during that interval.

        Note that sample times are determined without considering the
        actual parameter values. Two separate queries with equal start/stop
        arguments will always return the same number of samples with the
        same timestamps. This is done to ease merging of multiple sample
        series. You should always be explicit about the ``start`` and ``stop``
        times when relying on this property.

        :param parameter:
            Either a fully-qualified XTCE name or an alias in the
            format ``NAMESPACE/NAME``.
        :param start:
            Minimum generation time of the sampled
            parameter values (inclusive). If not set
            this defaults to one hour ago.
        :param stop:
            Maximum generation time of the sampled
            parameter values (exclusive). If not set
            this defaults to the current time.
        :param sample_count:
            The number of returned samples.
        :param parameter_cache:
            Specify the name of the processor who's
            parameter cache is merged with already
            archived values. To disable results from
            the parameter cache, set this to ``None``.
        :param source:
            Specify how to retrieve parameter values. By
            default this uses the ``ParameterArchive`` which
            is optimized for retrieval. For Yamcs instances
            that do not enable the ``ParameterArchive``, you can
            still get results by specifying ``replay`` as the
            source. Replay requests take longer to return because
            the data needs to be reprocessed.
        """
        path = f"/archive/{self._instance}/parameters{parameter}/samples"
        now = datetime.now(tz=timezone.utc)
        params = {
            "count": sample_count,
            "source": source,
            "start": to_isostring(now - timedelta(hours=1)),
            "stop": to_isostring(now),
        }
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        if parameter_cache:
            params["processor"] = parameter_cache
        else:
            params["norealtime"] = True

        response = self.ctx.get_proto(path=path, params=params)
        message = pvalue_pb2.TimeSeries()
        message.ParseFromString(response.content)
        samples = getattr(message, "sample")
        return [MeanSample(s) for s in samples]

    def list_parameter_ranges(
        self,
        parameter: str,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        min_gap: Optional[float] = None,
        max_gap: Optional[float] = None,
        min_range: Optional[float] = None,
        max_values: int = 100,
        parameter_cache: str = "realtime",
    ) -> List[ParameterRange]:
        """
        Returns parameter ranges between the specified start and stop time.

        Each range indicates an interval during which this parameter's
        value was uninterrupted and unchanged.

        Ranges are a good fit for retrieving the value of a parameter
        that does not change frequently. For example an on/off indicator
        or some operational status. Querying ranges will then induce
        much less overhead than manually processing the output of
        :meth:`list_parameter_values` would.

        The maximum number of returned ranges is limited to 500.

        :param parameter:
            Either a fully-qualified XTCE name or an alias in the
            format ``NAMESPACE/NAME``.
        :param start:
            Minimum generation time of the considered values (inclusive)
        :param stop:
            Maximum generation time of the considered values (exclusive)
        :param min_gap:
            Time in seconds. Any gap (detected based on parameter
            expiration) smaller than this will be ignored.
            However if the parameter changes value, the ranges
            will still be split.
        :param max_gap:
            Time in seconds. If the distance between two
            subsequent parameter values is bigger than
            this value (but smaller than the parameter
            expiration), then an artificial gap is
            created. This also applies if there is no
            expiration defined for the parameter.
        :param min_range:
            Time in seconds. Minimum duration of returned
            ranges. If multiple values occur within the
            range, the most frequent can be accessed using
            the ``entries`` property.
        :param max_values:
            Maximum number of unique values, tallied across the
            full requested range. Use this in combination with
            ``min_range`` to further optimize for transfer size.
            This value is limited to 100 at most.
        :param parameter_cache:
            Specify the name of the processor who's
            parameter cache is merged with already
            archived values. To disable results from
            the parameter cache, set this to ``None``.
        """
        path = f"/archive/{self._instance}/parameters{parameter}/ranges"
        params = {}
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if min_gap is not None:
            params["minGap"] = int(min_gap * 1000)
        if max_gap is not None:
            params["maxGap"] = int(max_gap * 1000)
        if min_range is not None:
            params["minRange"] = int(min_range * 1000)
        if max_values is not None:
            params["maxValues"] = max_values

        if parameter_cache:
            params["processor"] = parameter_cache
        else:
            params["norealtime"] = True

        response = self.ctx.get_proto(path=path, params=params)
        message = pvalue_pb2.Ranges()
        message.ParseFromString(response.content)
        ranges = getattr(message, "range")
        return [ParameterRange(r) for r in ranges]

    def list_parameter_values(
        self,
        parameter: str,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        page_size: int = 500,
        descending: bool = False,
        parameter_cache: str = "realtime",
        source: str = "ParameterArchive",
    ) -> Iterable[ParameterValue]:
        """
        Reads parameter values between the specified start and stop time.

        .. note::

            This method will send out multiple requests when more than
            ``page_size`` values are queried. For large queries, consider
            using :meth:`stream_parameter_values` instead, it uses server-streaming
            based on a single request, and supports downloading the values of
            multiple parameter at the same time.

        :param parameter:
            Either a fully-qualified XTCE name or an alias in the
            format ``NAMESPACE/NAME``.
        :param start:
            Minimum generation time of the returned values (inclusive)
        :param stop:
            Maximum generation time of the returned values (exclusive)
        :param page_size:
            Page size of underlying requests. Higher values imply
            less overhead, but risk hitting the maximum message size
            limit.
        :param descending:
            If set to ``True`` values are fetched in reverse
            order (most recent first).
        :param parameter_cache:
            Specify the name of the processor who's
            parameter cache is merged with already
            archived values. To disable results from
            the parameter cache, set this to ``None``.
        :param source:
            Specify how to retrieve parameter values. By
            default this uses the ``ParameterArchive`` which
            is optimized for retrieval. For Yamcs instances
            that do not enable the ``ParameterArchive``, you can
            still get results by specifying ``replay`` as the
            source. Replay requests take longer to return because
            the data needs to be reprocessed.
        """
        params: Dict[str, Any] = {
            "source": source,
            "order": "desc" if descending else "asc",
        }
        if page_size is not None:
            params["limit"] = page_size
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        if parameter_cache:
            params["processor"] = parameter_cache
        else:
            params["norealtime"] = True

        encoded_name = urllib.parse.quote_plus(parameter)
        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/parameters/{encoded_name}",
            params=params,
            response_class=pvalue_service_pb2.ListParameterHistoryResponse,
            items_key="parameter",
            item_mapper=ParameterValue,
        )

    def stream_parameter_values(
        self,
        parameters: Union[str, List[str]],
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        tm_links: Optional[Union[str, List[str]]] = None,
        chunk_size: int = 32 * 1024,
    ) -> Iterable[ParameterData]:
        """
        Reads parameter values between the specified start and stop time.

        Value updates are emitted for each unique generation time within
        the queried range. If one of the parameters does not have a value
        for a specific generation time, it is not included in the update.

        :param parameters:
            Parameter(s) to be queried.
        :param start:
            Minimum generation time of the returned values (inclusive)
        :param stop:
            Maximum generation time of the returned values (exclusive)
        :param tm_links:
            If set, include only values that were received through
            a specific link.
        """
        options = pvalue_service_pb2.StreamParameterValuesRequest()
        options.ids.extend(to_named_object_ids(parameters))
        if start is not None:
            options.start.MergeFrom(to_server_time(start))
        if stop is not None:
            options.stop.MergeFrom(to_server_time(stop))
        if tm_links:
            if isinstance(tm_links, str):
                options.tmLinks.extend([tm_links])
            else:
                options.tmLinks.extend(tm_links)

        def generate():
            path = f"/stream-archive/{self._instance}:streamParameterValues"
            response = self.ctx.post_proto(
                path=path, data=options.SerializeToString(), stream=True
            )
            for message in split_protobuf_stream(
                response.iter_content(chunk_size=chunk_size), pvalue_pb2.ParameterData
            ):
                yield ParameterData(message)

        return generate()

    def list_command_history(
        self,
        command: Optional[str] = None,
        queue: Optional[str] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        page_size: int = 500,
        descending: bool = False,
    ) -> Iterable[CommandHistory]:
        """
        Reads command history entries between the specified start and stop time.

        .. note::

            This method will send out multiple requests when more than
            ``page_size`` commands are queried. For large queries, consider
            using :meth:`stream_command_history` instead, it uses server-streaming
            based on a single request.

        :param command:
            Either a fully-qualified XTCE name or an alias in the
            format ``NAMESPACE/NAME``.
        :param queue:
            Name of the queue that the command was assigned to.
        :param start:
            Minimum generation time of the returned
            command history entries (inclusive)
        :param stop:
            Maximum generation time of the returned
            command history entries (exclusive)
        :param page_size:
            Page size of underlying requests. Higher values imply
            less overhead, but risk hitting the maximum message size
            limit.
        :param descending:
            If set to ``True`` results are fetched in reverse
            order (most recent first).
        """
        params: Dict[str, Any] = {
            "order": "desc" if descending else "asc",
        }
        if queue:
            params["queue"] = queue
        if page_size is not None:
            params["limit"] = page_size
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        if command:
            path = f"/archive/{self._instance}/commands{command}"
        else:
            path = f"/archive/{self._instance}/commands"

        return pagination.Iterator(
            ctx=self.ctx,
            path=path,
            params=params,
            response_class=commands_service_pb2.ListCommandsResponse,
            items_key="entry",
            item_mapper=CommandHistory,
        )

    def stream_command_history(
        self,
        name: Optional[Union[str, List[str]]] = None,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
        chunk_size: int = 32 * 1024,
    ) -> Iterable[CommandHistory]:
        """
        Reads command history between the specified start and stop time.

        .. versionadded:: 1.9.1

        :param name:
            If specified, return only commands with this name.
        :param start:
            Minimum generation time of the returned commands (inclusive)
        :param stop:
            Maximum generation time of the returned commands (exclusive)
        """
        options = commands_service_pb2.StreamCommandsRequest()
        if name is not None:
            if isinstance(name, str):
                options.name.extend([name])
            else:
                options.name.extend(name)
        if start is not None:
            options.start.MergeFrom(to_server_time(start))
        if stop is not None:
            options.stop.MergeFrom(to_server_time(stop))

        def generate():
            path = f"/stream-archive/{self._instance}:streamCommands"
            response = self.ctx.post_proto(
                path=path, data=options.SerializeToString(), stream=True
            )
            for message in split_protobuf_stream(
                response.iter_content(chunk_size=chunk_size),
                commanding_pb2.CommandHistoryEntry,
            ):
                yield CommandHistory(message)

        return generate()

    def list_tables(self) -> Iterable[Table]:
        """
        Returns the existing tables.

        Tables are returned in lexicographical order.
        """
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        path = f"/archive/{self._instance}/tables"
        response = self.ctx.get_proto(path=path)
        message = table_pb2.ListTablesResponse()
        message.ParseFromString(response.content)
        tables = getattr(message, "tables")
        return iter([Table(table) for table in tables])

    def get_table(self, table: str) -> Table:
        """
        Gets a single table.

        :param table:
            The name of the table.
        """
        path = f"/archive/{self._instance}/tables/{table}"
        response = self.ctx.get_proto(path=path)
        message = table_pb2.TableInfo()
        message.ParseFromString(response.content)
        return Table(message)

    def dump_table(
        self, table: str, query: Optional[str] = None, chunk_size: int = 32 * 1024
    ):
        path = f"/archive/{self._instance}/tables/{table}:readRows"
        if query:
            req = table_pb2.ReadRowsRequest()
            req.query = query
            response = self.ctx.post_proto(
                path=path, stream=True, data=req.SerializeToString()
            )
        else:
            response = self.ctx.post_proto(path=path, stream=True)
        return response.iter_content(chunk_size=chunk_size)

    def load_table(self, table: str, data, chunk_size: int = 32 * 1024) -> int:
        def read_in_chunks(file_object, chunk_size):
            chunk = file_object.read(chunk_size)
            while chunk:
                yield chunk
                chunk = file_object.read(chunk_size)

        path = f"/archive/{self._instance}/tables/{table}:writeRows"
        if chunk_size:
            generator = read_in_chunks(data, chunk_size)
            response = self.ctx.post_proto(path=path, data=generator)
        else:
            response = self.ctx.post_proto(path=path, data=data)
        message = table_pb2.WriteRowsResponse()
        message.ParseFromString(response.content)
        return message.count

    def rebuild_histogram(
        self,
        table: str,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
    ):
        """
        Rebuilds the histogram for a table. This may be necessary
        for example after bulk loading data.

        The rebuild may be constrained by using the
        ``start`` and ``stop`` parameters. When
        specified all partitions overlapping this range
        are reconsidered.

        .. note::
            Histogram rebuilds run synchronously: this
            method will await the outcome.

        :param table:
            The name of the table
        :param start:
            Start time
        :param stop:
            Stop time
        """
        req = table_pb2.RebuildHistogramRequest()
        if start:
            req.start.MergeFrom(to_server_time(start))
        if stop:
            req.stop.MergeFrom(to_server_time(stop))
        url = f"/archive/{self._instance}/tables/{table}:rebuildHistogram"
        self.ctx.post_proto(url, data=req.SerializeToString())

    def rebuild_parameter_archive(
        self,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
    ):
        """
        Rebuilds the Parameter Archive.

        The rebuild may be constrained by using the
        ``start`` and ``stop`` parameters. These values
        are only hints to the Parameter Archive, which
        will extend the requested range based on archive
        segmentation.

        .. note::
            Rebuilds run as an asynchronous operation: this
            method will not await the outcome.

        :param start:
            Start time. This argument is optional since Yamcs v5.11.4
        :param stop:
            Stop time. This argument is optional since Yamcs v5.11.4
        """
        req = parameter_archive_service_pb2.RebuildRangeRequest()
        if start:
            req.start.MergeFrom(to_server_time(start))
        if stop:
            req.stop.MergeFrom(to_server_time(stop))
        url = f"/archive/{self._instance}/parameterArchive:rebuild"
        self.ctx.post_proto(url, data=req.SerializeToString())

    def purge_parameter_archive(self):
        """
        Removes all Parameter Archive data and related metadata.

        Afterwards, use :meth:`rebuild_parameter_archive` to rebuild the Parameter
        Archive.
        """
        req = parameter_archive_service_pb2.PurgeRequest()
        url = f"/archive/{self._instance}/parameterArchive:purge"
        self.ctx.post_proto(url, data=req.SerializeToString())

    def enable_parameter_archive_backfilling(self):
        """
        Enables the automatic backfilling (rebuilding) of the parameter archive.

        If the backfilling is already enabled, this operation has no effect.
        """
        req = parameter_archive_service_pb2.EnableBackfillingRequest()
        url = f"/archive/{self._instance}/parameterArchive:enableBackfilling"
        self.ctx.post_proto(url, data=req.SerializeToString())

    def disable_parameter_archive_backfilling(self):
        """
        Disables the automatic backfilling (rebuilding) of the parameter archive.

        If the backfilling is already disabled, this operation has no effect.
        """
        req = parameter_archive_service_pb2.DisableBackfillingRequest()
        url = f"/archive/{self._instance}/parameterArchive:disableBackfilling"
        self.ctx.post_proto(url, data=req.SerializeToString())

    def rebuild_ccsds_index(
        self,
        start: Optional[datetime] = None,
        stop: Optional[datetime] = None,
    ):
        """
        Rebuilds the CCSDS index. This is only applicable to projects that use a
        CcsdsTmIndex service to calculate packet completeness.

        .. note::
            Index rebuilds run synchronously: this
            method will await the outcome.

        :param start:
            Start time
        :param stop:
            Stop time
        """
        req = index_service_pb2.RebuildCcsdsIndexRequest()
        if start:
            req.start.MergeFrom(to_server_time(start))
        if stop:
            req.stop.MergeFrom(to_server_time(stop))
        url = f"/archive/{self._instance}:rebuildCcsdsIndex"
        self.ctx.post_proto(url, data=req.SerializeToString())

    def list_streams(self) -> Iterable[Stream]:
        """
        Returns the existing streams.

        Streams are returned in lexicographical order.
        """
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        path = f"/archive/{self._instance}/streams"
        response = self.ctx.get_proto(path=path)
        message = table_pb2.ListStreamsResponse()
        message.ParseFromString(response.content)
        streams = getattr(message, "streams")
        return iter([Stream(stream) for stream in streams])

    def get_stream(self, stream: str) -> Stream:
        """
        Gets a single stream.

        :param stream:
            The name of the stream.
        """
        path = f"/archive/{self._instance}/streams/{stream}"
        response = self.ctx.get_proto(path=path)
        message = table_pb2.StreamInfo()
        message.ParseFromString(response.content)
        return Stream(message)

    def create_stream_subscription(
        self, stream: str, on_data: Callable[[StreamData], None], timeout: float = 60
    ) -> WebSocketSubscriptionFuture:
        """
        Create a new stream subscription.

        :param stream:
            The name of the stream.
        :param on_data:
            Function that gets called with  :class:`.StreamData`
            updates.
        :param timeout:
            The amount of seconds to wait for the request
            to complete.
        :return:
            Future that can be used to manage the background websocket
            subscription.
        """
        options = table_pb2.SubscribeStreamRequest()
        options.instance = self._instance
        options.stream = stream

        manager = WebSocketSubscriptionManager(
            self.ctx, topic="stream", options=options
        )

        # Represent subscription as a future
        subscription = WebSocketSubscriptionFuture(manager)

        wrapped_callback = functools.partial(
            _wrap_callback_parse_stream_data, subscription, on_data
        )

        manager.open(wrapped_callback)

        # Wait until a reply or exception is received
        subscription.reply(timeout=timeout)

        return subscription

    def execute_sql(self, statement: str) -> ResultSet:
        """
        Executes a single SQL statement.

        :param statement:
            SQL string
        :return:
            A result set for consuming rows
        """
        path = f"/archive/{self._instance}:executeStreamingSql"
        req = table_pb2.ExecuteSqlRequest()
        req.statement = statement

        response = self.ctx.post_proto(
            path=path, data=req.SerializeToString(), stream=True
        )
        return ResultSet(response)



import binascii
import collections
import json
import logging
import os
from datetime import datetime, timezone
from threading import Timer
from typing import Any, Callable, Iterator, List, Union
from urllib.parse import urlparse

import requests
import urllib3
from google.protobuf import timestamp_pb2
from google.protobuf.internal.decoder import _DecodeVarint32
from google.protobuf.internal.encoder import _VarintBytes
from yamcs.client.core.exceptions import ConnectionFailure, YamcsError
from yamcs.protobuf import yamcs_pb2

__all__ = [
    "adapt_name_for_rest",
    "delimit_protobuf",
    "do_get",
    "do_post",
    "do_request",
    "FixedDelay",
    "parse_server_time",
    "parse_server_timestring",
    "parse_value",
    "ProtoList",
    "split_protobuf_stream",
    "to_argument_value",
    "to_isostring",
    "to_named_object_id",
    "to_named_object_ids",
    "to_server_time",
]

logger = logging.getLogger("yamcs-client")


def to_isostring(dt: datetime) -> str:
    """
    Converts the given datetime to an ISO String for use as URL parameter.
    """
    pb = to_server_time(dt)
    return pb.ToJsonString()


def to_server_time(dt: datetime) -> timestamp_pb2.Timestamp:
    """
    Converts the given ``datetime.datetime`` to a
    ``google.protobuf.timestamp_pb2.Timestamp``.
    """
    if dt.tzinfo is None:
        # TODO - this is a temporary error.
        # Old versions of this client interpreted naive times as UTC,
        # but we would like to gradually adapt to use the system's
        # default instead. Until known users have adapted, the
        # safest solution is to just enforce a timezone.
        raise YamcsError(
            "Datetimes must be timezone-aware. For example: use "
            + "``datetime.now(tz=timezone.utc)`` instead of "
            + "``datetime.now()``"
        )

    pb = timestamp_pb2.Timestamp()
    pb.FromDatetime(dt)
    return pb


def parse_server_timestring(isostring: str) -> datetime:
    """
    Converts an ISO string to a timezone-aware ``datetime.datetime``.
    The timezone uses the system-default. This can be overriden to UTC
    by setting the environment variable ``PYTHON_YAMCS_CLIENT_UTC``.
    """
    isostring = isostring.replace("Z", "GMT")
    try:
        naive = datetime.strptime(isostring, "%Y-%m-%dT%H:%M:%S.%f%Z")
    except ValueError:
        # Protobuf's ToJsonString does not emit fractional digits if 0
        naive = datetime.strptime(isostring, "%Y-%m-%dT%H:%M:%S%Z")
    utctime = naive.replace(tzinfo=timezone.utc)

    if os.environ.get("PYTHON_YAMCS_CLIENT_UTC") not in (None, "0"):
        return utctime
    return utctime.astimezone(tz=None)


def parse_server_time(pb: timestamp_pb2.Timestamp) -> datetime:
    """
    Converts a Protobuf timestamp message to a timezone-aware ``datetime.datetime``.
    The timezone uses the system-default. This can be overriden to UTC by setting
    the environment variable ``PYTHON_YAMCS_CLIENT_UTC``.
    """
    utctime = pb.ToDatetime().replace(tzinfo=timezone.utc)

    if os.environ.get("PYTHON_YAMCS_CLIENT_UTC") not in (None, "0"):
        return utctime
    return utctime.astimezone(tz=None)


def parse_value(proto: yamcs_pb2.Value) -> Any:
    """
    Converts a Protobuf `Value` from the API into a python native value
    """
    if proto.type == yamcs_pb2.Value.FLOAT:
        return proto.floatValue
    elif proto.type == yamcs_pb2.Value.DOUBLE:
        return proto.doubleValue
    elif proto.type == yamcs_pb2.Value.SINT32:
        return proto.sint32Value
    elif proto.type == yamcs_pb2.Value.UINT32:
        return proto.uint32Value
    elif proto.type == yamcs_pb2.Value.BINARY:
        return proto.binaryValue
    elif proto.type == yamcs_pb2.Value.TIMESTAMP:
        # Don't use the actual 'timestampValue' field, it contains a number
        # that is difficult to interpret on the client. Instead parse from
        # the ISO String also set by Yamcs.
        return parse_server_timestring(proto.stringValue)
    elif proto.type == yamcs_pb2.Value.STRING:
        return proto.stringValue
    elif proto.type == yamcs_pb2.Value.UINT64:
        return proto.uint64Value
    elif proto.type == yamcs_pb2.Value.SINT64:
        return proto.sint64Value
    elif proto.type == yamcs_pb2.Value.BOOLEAN:
        return proto.booleanValue
    elif proto.type == yamcs_pb2.Value.ENUMERATED:
        return proto.stringValue
    elif proto.type == yamcs_pb2.Value.ARRAY:
        return [parse_value(v) for v in proto.arrayValue]
    elif proto.type == yamcs_pb2.Value.AGGREGATE:
        tuples = zip(proto.aggregateValue.name, proto.aggregateValue.value)
        return dict(map(lambda x: (x[0], parse_value(x[1])), tuples))
    elif proto.type == yamcs_pb2.Value.NONE:
        return None
    else:
        logger.warning("Unrecognized value type for update %s", proto)
        return None


def adapt_name_for_rest(name: str) -> str:
    """
    Modifies a user-provided name for use in API calls.
    Basically we want an alias like 'MDB:OPS Name/SIMULATOR_BatteryVoltage2'
    to be prepended with a slash.
    """
    if name.startswith("/"):
        return name
    elif "/" not in name:
        raise ValueError("Provided name is not a fully-qualified XTCE name.")
    return "/" + name


def to_named_object_id(parameter: str) -> yamcs_pb2.NamedObjectId:
    """
    Builds a NamedObjectId. This is a bit more complex than it really
    should be. In Python (for convenience) we allow the user to simply address
    entries by their alias via the NAMESPACE/NAME convention. Yamcs is not
    aware of this convention so we decompose it into distinct namespace and
    name fields.
    """
    named_object_id = yamcs_pb2.NamedObjectId()
    if parameter.startswith("/"):
        named_object_id.name = parameter
    else:
        parts = parameter.split("/", 1)
        if len(parts) < 2:
            raise ValueError(
                f"Failed to process {parameter}. Use fully-qualified "
                "XTCE names or, alternatively, an alias in "
                "in the format NAMESPACE/NAME"
            )
        named_object_id.namespace = parts[0]
        named_object_id.name = parts[1]
    return named_object_id


def to_named_object_ids(
    parameters: Union[str, List[str]]
) -> List[yamcs_pb2.NamedObjectId]:
    """Builds a list of NamedObjectId."""
    if isinstance(parameters, str):
        return [to_named_object_id(parameters)]
    return [to_named_object_id(parameter) for parameter in parameters]


def to_argument_value(value, force_string):
    if isinstance(value, (bytes, bytearray)):
        return binascii.hexlify(value).decode("ascii")
    elif isinstance(value, str):
        return value
    elif isinstance(value, collections.abc.Mapping):
        # Careful to do the JSON dump only at the end,
        # and not at every level of a nested hierarchy
        obj = _compose_aggregate_members(value)
        return json.dumps(obj)
    elif isinstance(value, collections.abc.Sequence):
        # Yamcs expects a JSON encoded array, where elements are
        # themselves JSON-encoded (to be improved in a future version).
        obj = [to_argument_value(x, force_string=True) for x in value]
        return json.dumps(obj)
    elif isinstance(value, datetime):
        return to_isostring(value)
    elif force_string:
        return str(value)
    else:
        return value


def _compose_aggregate_members(value):
    """
    Recursively creates an object that can eventually be serialized to a valid
    aggregate value in JSON. This is a bit different than non-aggregate values,
    because Yamcs is more strict in the values that it accepts (for example:
    unlike regular arguments you cannot assign a numeric string to an integer
    argument, the JSON type needs to be numeric too).
    """
    if isinstance(value, (bytes, bytearray)):
        return binascii.hexlify(value).decode("ascii")
    elif isinstance(value, collections.abc.Mapping):
        return {k: _compose_aggregate_members(v) for k, v in value.items()}
    elif isinstance(value, datetime):
        return to_isostring(value)
    else:
        # No string conversion here, use whatever the user is giving
        return value


def do_get(session: requests.Session, path: str, **kwargs) -> requests.Response:
    """
    Performs an HTTP GET request while reraising connection-type exceptions
    to something produced by this library.
    """
    return do_request(session, "get", path, **kwargs)


def do_post(session: requests.Session, path: str, **kwargs) -> requests.Response:
    """
    Performs an HTTP POST request while reraising connection-type exceptions
    to something produced by this library.
    """
    return do_request(session, "post", path, **kwargs)


def do_request(
    session: requests.Session, method: str, path: str, **kwargs
) -> requests.Response:
    """
    Performs an HTTP request while reraising connection-type exceptions
    to something produced by this library.
    """
    try:
        session.request
        return session.request(method, path, **kwargs)
    except requests.exceptions.SSLError as ssl_error:
        url_parts = urlparse(path)
        base_url = f"{url_parts.scheme}://{url_parts.netloc}"
        msg = f"Connection to {base_url} failed: {ssl_error}"
        raise ConnectionFailure(msg) from None
    except requests.exceptions.ConnectionError as e:
        url_parts = urlparse(path)
        base_url = f"{url_parts.scheme}://{url_parts.netloc}"

        # Requests gives us a very confusing error when a connection
        # is refused. Confirm and unwrap.
        if e.args and isinstance(e.args[0], urllib3.exceptions.MaxRetryError):
            # This is a string (which is still confusing ....)
            msg = e.args[0].args[0]
            if "refused" in msg:
                msg = f"Connection to {base_url} failed: connection refused"
                raise ConnectionFailure(msg) from None
            elif "not known" in msg:
                msg = f"Connection to {base_url} failed: could not resolve hostname"
                raise ConnectionFailure(msg) from None

        raise ConnectionFailure(f"Connection to {base_url} failed: {e}")


def delimit_protobuf(message_generator: Iterator[Any], chunk_size: int = 32 * 1024):
    buf = None
    for message in message_generator:
        msg_buf = _VarintBytes(message.ByteSize()) + message.SerializeToString()
        if buf is None:
            buf = msg_buf
        else:
            buf += msg_buf
        # Not exact, use the chunk size as a treshold
        if len(buf) >= chunk_size:
            yield buf
            buf = None

    if buf:
        yield buf


def split_protobuf_stream(chunk_iterator, message_class):
    buf = None
    for chunk in chunk_iterator:
        if buf is None:
            buf = chunk
        else:
            buf += chunk

        while len(buf):
            try:
                # n is advanced beyond the varint
                msg_len, n = _DecodeVarint32(buf, 0)
            except IndexError:
                break  # Need another chunk

            if n + msg_len > len(buf):
                break  # Need another chunk

            msg_buf = buf[n : (n + msg_len)]
            buf = buf[(n + msg_len) :]
            message = message_class()
            message.ParseFromString(msg_buf)
            yield message


class ProtoList(list):
    """
    List subclass to work with a list of wrapped items while
    maintaining the state in an underlying protobuf message.

    This assumes each wrapped item has an internal `_proto`
    field.
    """

    def __init__(self, parent_proto, items_key, item_mapper):
        self.parent_proto = parent_proto
        self.items_key = items_key
        self.item_mapper = item_mapper

    def __getitem__(self, index):
        repeatable = getattr(self.parent_proto, self.items_key)
        proto = repeatable.__getitem__(index)
        return self.item_mapper(proto)

    def __setitem__(self, index, value):
        repeatable = getattr(self.parent_proto, self.items_key)
        repeatable.__setitem__(index, value._proto)

    def __delitem__(self, __i):
        repeatable = getattr(self.parent_proto, self.items_key)
        return repeatable.__delitem__(__i)

    def __len__(self):
        repeatable = getattr(self.parent_proto, self.items_key)
        return repeatable.__len__()

    def __iter__(self):
        repeatable = getattr(self.parent_proto, self.items_key)
        return [self.item_mapper(proto) for proto in repeatable].__iter__()

    def __add__(self, __x):
        # Not obvious what should be the message_listener
        # of the returned repeated composite field container.
        raise TypeError(
            f"{self.__class__.__name__} object does not support list addition"
        )

    def __iadd__(self, __x):
        repeatable = getattr(self.parent_proto, self.items_key)
        protos = [item._proto for item in __x]
        repeatable.extend(protos)
        return self

    def reverse(self):
        repeatable = getattr(self.parent_proto, self.items_key)
        try:
            repeatable.reverse()
        except AttributeError:  # Only available since protobuf>=3.15
            raise TypeError(
                f"{self.__class__.__name__} object does not support reverse operation"
            ) from None

    def append(self, value):
        repeatable = getattr(self.parent_proto, self.items_key)
        repeatable.append(value._proto)

    def extend(self, __iterable):
        repeatable = getattr(self.parent_proto, self.items_key)
        repeatable.extend([value._proto for value in __iterable])

    def clear(self):
        repeatable = getattr(self.parent_proto, self.items_key)
        copy = list(repeatable)
        for item in copy:
            repeatable.remove(item)


class FixedDelay:
    """
    Helper class to run a periodic action, with a fixed delay between
    the termination of one execution, and the commencement of the next.
    """

    def __init__(
        self,
        action: Callable[[], None],
        initial_delay: float,
        period: float,
    ):
        self._timer = None
        self.action = action
        self.period = period
        self.is_running = False
        self.start(initial_delay)

    def start(self, period):
        if not self.is_running:
            self._timer = Timer(period, self._run)
            self._timer.daemon = True
            self._timer.start()
            self.is_running = True

    def stop(self):
        if self._timer:
            self._timer.cancel()
        self.is_running = False

    def _run(self):
        self.is_running = False
        self.start(self.period)
        self.action()
